# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import socket
from typing import Optional

from tikka.adapters.network.thread_safe_substrate_interface import (
    ThreadSafeSubstrateInterface,
)
from tikka.domains.entities.node import Node
from tikka.interfaces.adapters.network.connection import ConnectionInterface

# websocket timeout
RPC_CONNECTION_TIMEOUT = 30


class RPCConnection(ConnectionInterface):
    """
    RPCConnection class
    """

    def __init__(self) -> None:
        """
        Init RPCConnection instance

        RPC client is available in self.client after connect()
        """
        self.client: Optional[ThreadSafeSubstrateInterface] = None

    def connect(self, node: Node) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.connect.__doc__
        )
        try:
            # fixme:    use type_registry_preset='default' to fix error on Type MultiAddress
            #           https://github.com/polkascan/py-substrate-interface/issues/110
            self.client = ThreadSafeSubstrateInterface(
                url=node.url,
                ws_options={
                    "timeout": RPC_CONNECTION_TIMEOUT,
                    "sockopt": ((socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),),
                },
            )
        except Exception as exception:
            self.client = None
            logging.exception(exception)

        if self.client is not None:
            # fixme: workaround for a decode error with fees see #7
            self.client.config["rpc_methods"] = self.client.rpc_request(
                "rpc_methods", []
            )["result"]["methods"]
            self.client.config["rpc_methods"].remove("state_call")

    def disconnect(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.disconnect.__doc__
        )
        if self.client is not None:
            self.client.close()
            self.client = None

    def is_connected(self) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.is_connected.__doc__
        )
        return self.client is not None



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pathlib import Path
from typing import Optional, Union

from duniterpy.key import SigningKey

from tikka.domains.entities.wallet import V1FileWallet
from tikka.interfaces.adapters.repository.file_wallets import (
    V1FileWalletsRepositoryInterface,
)


class V1FileWalletsRepository(V1FileWalletsRepositoryInterface):
    """
    V1FileWalletsRepository class
    """

    def load(
        self, path: Union[str, Path], password: Optional[str] = None
    ) -> V1FileWallet:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            V1FileWalletsRepositoryInterface.load.__doc__
        )
        if isinstance(path, str):
            path = Path(path)

        type_ = self.get_type(path)
        is_encrypted = type_ == "EWIF"
        signing_key: Optional[SigningKey] = None
        if not is_encrypted:
            if type_ == "WIF":
                signing_key = SigningKey.from_wif_file(str(path))
            elif type_ == "PUBSEC":
                signing_key = SigningKey.from_pubsec_file(str(path))
        elif password is not None:
            if type_ == "EWIF":
                signing_key = SigningKey.from_ewif_file(
                    str(path), password
                )  # type: ignore

        return V1FileWallet(path, type_, is_encrypted, signing_key)

    @staticmethod
    def get_type(path: Union[str, Path]) -> str:
        """
        Get format type of the V1 wallet file

        :param path: Path instance or string of the file
        :return:
        """
        if isinstance(path, str):
            path = Path(path)
        if path.suffix == ".dunikey":
            with path.open("r", encoding="utf-8") as file_handle:
                _type = file_handle.readline().strip()

                if "Type: EWIF" in _type:
                    type_ = "EWIF"
                elif "Type: WIF" in _type:
                    type_ = "WIF"
                elif "Type: PubSec" in _type:
                    type_ = "PUBSEC"

        return type_



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from datetime import datetime
from typing import List, Optional

from substrateinterface import Keypair

from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.interfaces.adapters.network.smiths import NetworkSmithsInterface
from tikka.interfaces.adapters.repository.smiths import SmithsRepositoryInterface


class Smiths:

    """
    Smiths domain class
    """

    def __init__(
        self,
        repository: SmithsRepositoryInterface,
        network: NetworkSmithsInterface,
    ):
        """
        Init Smiths domain

        :param repository: SmithsRepositoryInterface instance
        :param network: NetworkSmithsInterface instance
        """
        self.repository = repository
        self.network = network

    @staticmethod
    def create(
        identity_index: int,
        expire_on: Optional[datetime] = None,
        status: SmithStatus = SmithStatus.INVITED,
    ):
        """
        Return a Smith instance from params

        :param identity_index: Identity index number in blockchain
        :param expire_on: Smith status expiration timestamp
        :param status: Smith status
        :return:
        """
        return Smith(
            identity_index=identity_index,
            expire_on=expire_on,
            status=status,
        )

    def add(self, smith: Smith):
        """
        Add smith in repository

        :param smith: Smith instance
        :return:
        """
        self.repository.add(smith)

    def update(self, smith: Smith):
        """
        Update smith in repository

        :param smith: Smith instance
        :return:
        """
        self.repository.update(smith)

    def get(self, identity_index: int) -> Optional[Smith]:
        """
        Get Smith instance from Identity index

        :param identity_index: Identity index
        :return:
        """
        return self.repository.get(identity_index)

    def delete(self, identity_index: int) -> None:
        """
        Delete smith in repository

        :param identity_index: Identity index to delete
        :return:
        """
        self.repository.delete(identity_index)

    def exists(self, identity_index: int) -> bool:
        """
        Return True if smith exists in repository

        :param identity_index: Identity index to check
        :return:
        """
        return self.repository.exists(identity_index)

    def list(self) -> List[Smith]:
        """
        Return list of all smiths in repository

        :return:
        """
        return self.repository.list()

    def network_invite_member(self, keypair: Keypair, identity_index: int) -> None:
        """
        Request a smith membership for the Keypair account with node session_keys

        :param keypair: Owner Keypair
        :param identity_index: Identity index of member to invite to be smith
        :return:
        """
        return self.network.invite(keypair, identity_index)

    def network_accept_invitation(self, keypair: Keypair) -> None:
        """
        Invited Keypair account smith invitation acceptance

        :param keypair: Owner Keypair
        :return:
        """
        self.network.accept_invitation(keypair)

    def network_certify(self, keypair: Keypair, identity_index: int) -> None:
        """
        Certify an identity to be smith with the Keypair account

        :param keypair: Owner Keypair
        :param identity_index: Identity index of member to invite to be smith
        :return:
        """
        self.network.certify(keypair, identity_index)

    def network_get_smith(self, identity_index: int) -> Optional[Smith]:
        """
        Get smith for identity index from network if any

        :param identity_index: Identity index
        :return:
        """
        smith = self.network.get_smith(identity_index)
        if smith is not None:
            if self.exists(identity_index) is True:
                self.update(smith)
            else:
                self.add(smith)
        else:
            self.delete(identity_index)

        return smith

    def network_get_smiths(self, identity_indice: List[int]) -> None:
        """
        Get Smith instances from network from identity_indice list

        :param identity_indice: Identity indice list
        :return:
        """
        smiths = self.network.get_smiths(identity_indice)
        for index, smith in enumerate(smiths):
            if smith is None:
                self.repository.delete(identity_indice[index])
                continue
            if self.exists(smith.identity_index) is True:
                self.update(smith)
            else:
                self.add(smith)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import Enum
from typing import TypeVar

AuthorityType = TypeVar("AuthorityType", bound="Authority")


class AuthorityStatus(int, Enum):
    OFFLINE = 0
    ONLINE = 1
    INCOMING = 2
    OUTGOING = 3


@dataclass
class Authority:
    identity_index: int
    status: AuthorityStatus
